



capnp --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
capnpc --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
capnpc-c++ --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
capnpc-capnp --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_INC%\\capnp exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_LIB%\\capnp.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %LIBRARY_LIB%\\kj.lib exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
